#include <cmath>
#include <cstdio>
#include <string>
#include <vector>

extern "C" { 
int LoggingEnabled; /* 0=no, 1=yes, 2=recording! */
}

/* Note: This module assumes everyone uses RGB15 as display depth */

static std::string VIDEO_CMD =
    "mencoder - -o test0.avi"
    " -noskip -mc 0"
    " -ovc lavc"
    " -lavcopts vcodec=ffv1:context=0:format=BGR32:coder=1:vstrict=-1"
    " >& mencoder.log";
static std::string AUDIO_CMD =
    "cat > s.log";

static unsigned FPS = 60;

namespace AVI
{
    #define BGR24 0x42475218  // BGR24 fourcc
    #define BGR16 0x42475210  // BGR16 fourcc
    #define BGR15 0x4247520F  // BGR15 fourcc

    FILE* avifp = NULL;

    void AVIOpen()
    {
        avifp = popen(VIDEO_CMD.c_str(), "w");
    }

    void AVIClose()
    {
        pclose(avifp);
    }

    #define u32(n) (n)&255,((n)>>8)&255,((n)>>16)&255,((n)>>24)&255
    #define u16(n) (n)&255,((n)>>8)&255
    #define s4(s) s[0],s[1],s[2],s[3]

    void AVISend
        (const unsigned char *vidbuf,
         unsigned width, 
         unsigned height,
         bool first_frame)
    {
        const unsigned fourcc = BGR16;
        const unsigned framesize = width*height*2;
        
        if(first_frame)
        {
            const unsigned nframes = 0; //unknown
            const unsigned scale = 1000000;
            const unsigned scaled_fps = (unsigned)(scale * FPS);
            
            const unsigned SIZE_strh_vids = 4 + 4*2 + 2*2 + 8*4 + 2*4;
            const unsigned SIZE_strf_vids = 4*3 + 2*2 + 4*6;
            const unsigned SIZE_strl = 4+ 4+(4+SIZE_strh_vids) + 4+(4+SIZE_strf_vids);
            const unsigned SIZE_avih = 4*12;
            const unsigned SIZE_hdrl = 4+4+ (4+SIZE_avih) + 4 + (4+SIZE_strl);
            const unsigned SIZE_movi = 4 + nframes*(4+4+framesize);
            const unsigned SIZE_avi = 4+4+ (4+SIZE_hdrl) + 4 + (4+SIZE_movi);
            
            const unsigned char AVIheader[] =
            {
                s4("RIFF"),
                u32(SIZE_avi),
                s4("AVI "),   
                s4("LIST"),   
                u32(SIZE_hdrl),
                 s4("hdrl"),   
                 
                 s4("avih"),
                 u32(SIZE_avih),
                  u32(0),
                  u32(0),
                  u32(0),
                  u32(0),
                  u32(nframes),
                  u32(0),
                  u32(1), // one stream
                  u32(0),
                  u32(0),
                  u32(0),
                  u32(0),
                  u32(0),
                 
                 s4("LIST"),
                 
                 u32(SIZE_strl),
                  s4("strl"),   
                  
                   s4("strh"),
                   u32(SIZE_strh_vids),
                    s4("vids"),
                    u32(0),
                    u32(0),
                    u16(0),
                    u16(0),
                    u32(0),
                    u32(scale),
                    u32(scaled_fps),
                    u32(0),
                    u32(0),
                    u32(0),
                    u32(0),
                    u32(0),
                    u16(0),
                    u16(0),
                    u16(0),
                    u16(0),
                   
                   s4("strf"),
                   u32(SIZE_strf_vids),
                    u32(0),
                    u32(width),
                    u32(height),
                    u16(0),
                    u16(0),
                    u32(fourcc),
                    u32(0),
                    u32(0),
                    u32(0),
                    u32(0),
                    u32(0),
                s4("LIST"),
                u32(SIZE_movi),
                 s4("movi")
            };
              
            fwrite(AVIheader, 1, sizeof(AVIheader), avifp);
        }

        const unsigned char header[] = { s4("00dc"), u32(framesize) };
        
        fwrite(header, 1, sizeof(header), avifp);
        fwrite(vidbuf, 1, framesize, avifp);
    }
};

static class VideoLogging
{
    bool first;
public:
    VideoLogging(): first(true)
    {
        AVI::AVIOpen();
    }
    ~VideoLogging()
    {
        AVI::AVIClose();
    }
    void operator() (const void*data, unsigned width,unsigned height)
    {
        if(first) SendLogo(width,height);
        
        AVI::AVISend( (const unsigned char*)data, width, height, first);
        first=false;
    }
    void SendLogo(unsigned width, unsigned height)
    {
        /* Bisqwit's logo addition routine. */
        /* If you don't have his files, this function does nothing
         * and it does not matter at all.
         */
        
		const char *background =
			width==320 ? "logo320_240"
		  : width==512 ? "logo512_448"
		  : height>224 ? "logo256_240"
		  : "logo256_224";
		
		/* Note: This should be 1 second long. */
		for(unsigned frame = 0; frame < FPS; ++frame)
		{
		    char Buf[4096];
		    sprintf(Buf, "/shares/home/bisqwit/povray/nesvlogo/%s_f%u.tga",
		        background, frame);
			
			FILE*fp = fopen(Buf, "rb");
			if(!fp) continue; /* Silently ignore missing frames */
			
			int idlen = fgetc(fp);
			/* Silently ignore all other header data.
			 * These files are assumed to be uncompressed BGR24 tga files with Y swapped.
			 * Even their geometry is assumed to match perfectly.
			 */
			fseek(fp, 1+1+2+2+1+ /*org*/2+2+ /*geo*/2+2+ 1+1+idlen, SEEK_CUR);

			bool yflip=true;
			std::vector<unsigned char> data(width*height*3);
			for(unsigned y=height; y-->0; )
			    fread(&data[y*width*3], 1, width*3, fp);
			fclose(fp);
			
			std::vector<unsigned short> result(width*height);
			for(unsigned pos=0, max=result.size(); pos<max; ++pos)
			{
				unsigned usepos = pos;
				if(yflip)
				{
				    unsigned y = pos/width;
				    usepos = (usepos%width) + (height-y-1)*width;
				}
				
				unsigned B = data[usepos*3+0];
				unsigned G = data[usepos*3+1];
				unsigned R = data[usepos*3+2];
				result[pos] = ((B*31/255)<<0)
							| ((G*63/255)<<5)
							| ((R*31/255)<<11);
			}
			AVI::AVISend((const unsigned char*)&result[0], width,height, first);
			first=false;
		}
    }
} VideoLogging;

extern "C" {

const char* NESVideoGetVideoCmd()
{
    return VIDEO_CMD.c_str();
}
const char* NESVideoGetAudioCmd()
{
    return AUDIO_CMD.c_str();
}
void NESVideoSetVideoCmd(const char *cmd)
{
    VIDEO_CMD = cmd;
}
void NESVideoSetAudioCmd(const char *cmd)
{
    AUDIO_CMD = cmd;
}

void NESVideoLoggingVideo(const void*data, unsigned width,unsigned height)
{
    if(LoggingEnabled < 2) return;

    VideoLogging(data,width,height);
}

void NESVideoLoggingAudio(const void*data, unsigned bytes, unsigned bytes_per_second)
{
    if(LoggingEnabled < 2) return;

    static FILE* out = NULL;
    if(!out)
    {
        out = popen(AUDIO_CMD.c_str(), "w");
        if(out)
        {
            /* Insert 1 second of silence to the beginning */
            unsigned length = bytes_per_second;
            unsigned* EmptyBuffer = (unsigned*)malloc(length);
            fwrite(EmptyBuffer, 1, length, out);
            free(EmptyBuffer);
        }
    }
    if(out)
    {
        fwrite(data, 1, bytes, out);
        fflush(out);
    }
}

}
